/* CRCCHECK
 *
 * This daemon uses the "Stealth Bomber" package (donnated to the public domain
 * by Kevin Dean) to perform a few system and file integrity checks. We use
 * his library without modification.
 *
 * Although we provide the source and executable of this daemon, we do not
 * distribute any source or binaries of the "Stealth Bomber" (version 2.2)
 * package, except for the CRCSET program (this is explicitly permitted in the
 * license). The Stealth Bomber is available on CompuServe, Simtel and numerous
 * other BBSes.
 *
 * This daemon is non-standard in several ways. The most important difference
 * is that it does not use the dynamic linking features of the Daemon server;
 * that is, it never links to the Daemon server. On the other hand, it shows
 * another trick only possible with "Daemon-based" programs: terminating both
 * this program and the parent program (and thereby having Daemon server
 * cleaning up all other daemons).
 *
 * Compiled with Borland C++ 3.1, small memory model.
 */
#include <conio.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>

#include "vircheck.h"
#include "daemon.h"

/*
 * The Stealth Bomber library uses about 2 kBytes of stack. Therefore we set
 * the stack to 3 kBytes (we use very littly stack ourselves). For fastest
 * operation, we give the Stealth Bomber the maximum heap size it uses: 16
 * kBytes.
 *
 * If you use Microsoft C 5.1 or 6.0 (not sure about later versions), set the
 * stack size to 3 kBytes or more (the default is 2 kBytes, which is a very
 * critical size) with the LINK's /STACK: parameter.
 */
#if defined (__TURBOC__)
  extern unsigned _stklen = 3072;
  extern unsigned _heaplen = 17000;     /* large heap (faster) */
#endif


void exit_hard(int return_code);        /* prototype */

#define STEALTH_NFILES  2
const filecrc _fcrc[STEALTH_NFILES] =
{
  { '_', 'S', 'T', 'E', 'A', 'L', 'T', 'H' },
  { '_', 'D', 'A', 'E', 'M', 'O', 'N', '_' }
};

int main(int argc,char *argv[])
{
  int result,key;

  result = stealth_sys_check();
  if (result != STEALTH_OK) {
    fprintf(stderr, "\nDOS interrupt or memory subsystem is inconsistent (code %02x).\n"
                    "\nPress Enter to abort or Esc to continue", result);
    do
      if ((key = getch()) == 13)
        exit_hard(1);
    while (key != 13 && key != 27);
  } /* if */

  /* CRCCHECK checks DBIND. CRCSET modifies the _STEALTH signature for this
   * purpose. But the signature inside the CRCCHECK module should *not* be
   * changed, doing so makes the daemon useless for checking other program
   * files. So there is a separate signature for DBIND.
   */
  result = stealth_file_check(argv[0], _fcrc[0]);       /* first attempt */
  if (result != STEALTH_OK)
    result = stealth_file_check(argv[0], _fcrc[1]);     /* try DBIND's key */
  if ((result & STEALTH_NO_MEM) != 0)
    fprintf(stderr, "\nInsufficient memory for intergrity checking.\n");
  if ((result & STEALTH_FILE_ERR) != 0)
    fprintf(stderr, "\nUnable to read executable file.\n");
  if ((result & STEALTH_FILE_DATE_ERR) != 0
      || (result & STEALTH_FILE_SIZE_ERR) != 0)
    fprintf(stderr, "\nDOS File subsystem is inconsistent (code %02x).\n",
            result);
  if ((result & STEALTH_CRC_INVALID) != 0
      || (result & STEALTH_CRC_BAD_POLY) != 0)
    fprintf(stderr, "\nExecutable file damaged (code %02x).\n", result);

  if (result != STEALTH_OK) {
    fprintf(stderr, "\nPress Enter to abort or Esc to continue");
    do
      if ((key = getch()) == 13)
        exit_hard(1);
    while (key != 13 && key != 27);
  } /* if */

  return 0;     /* no consistency problems */
}

static unsigned get_psp(void)
{
  /* get the active PSP */
  union REGS regs;

  regs.h.ah = 0x62;
  int86(0x21,&regs,&regs);
  return regs.x.bx;
}

static void set_psp(unsigned psp)
{
  /* set the active PSP */
  union REGS regs;

  regs.h.ah = 0x50;                     /* previously undocumented */
  regs.x.bx = psp;
  int86(0x21,&regs,&regs);
}

static unsigned get_daemon(int daemon)
{
  /* get PSP address of installed daemon */
  union D_REGS d_regs;

  d_regs.x.ax = 0x0005;                 /* system function: get daemon addr. */
  d_regs.h.dl = daemon;
  d_call(&d_regs);
  return d_regs.x.ax;
}

static void set_daemon(int daemon, unsigned psp)
{
  /* set a PSP address at a daemon slot */
  union D_REGS d_regs;

  d_regs.x.ax = 0x0009;                 /* system function: set daemon addr. */
  d_regs.h.dl = daemon;
  d_regs.x.bx = psp;
  d_call(&d_regs);
}

void exit_hard(int return_code)
{
  int daemon;

  /* check that we are running under the daemon server */
  if (d_init() && d_inload()) {

    /*
     * When the application terminates, Daemon server takes control to unload
     * all daemons that were not explicitly unloaded by the application. But
     * the Daemon server only knows of daemons that have linked themselves to
     * it. This "daemon" never links. Now the dilema is:
     * o  if we end this daemon, the application continues to run (and we wish
     *    to abort it)
     * o  if we end the application, Daemon server cannot unload this daemon,
     *    so it is left as a sandbar in memory
     * Our solution is to patch this daemon it into the internal daemon table
     * of the Daemon server. This way the daemon would be known by the Daemon
     * server without having really been linked to it.
     * A different point we usually have to worry about in these circumstances
     * is the "inload" phase. Luckily, terminating everything works even in the
     * inload phase.
     */

    /* find an empty slot to patch our address in */
    for (daemon = 0; daemon < 256 && get_daemon(daemon) != 0; daemon++)
      /* nothing */;
    set_daemon(daemon, get_psp() );     /* patch in our PSP address */

    /* switch to the PSP address of the main application */
    set_psp( get_daemon(0) );

  } else {

    fprintf(stderr, "\nDaemon server absent\n");

  } /* if */

  /* now acting as the application, terminate */
  exit(return_code);
}

