/* UULIB
 *
 * This example program shows how to create a daemon that exports functions to
 * other applications.
 *
 * This daemon can encode a stream of binary data into ASCII strings that can
 * be transmitted over all networks that support ASCII. This encoding is called
 * "UU-encoding".
 *
 * The basic scheme is to break groups of 3 eight bit characters (24 bits) into
 * 4 six bit characters and then add 32 (a space) to each six bit character
 * which maps it into the readily transmittable character. As some transmission
 * mechanisms compress or remove spaces, spaces are changed into back-quote
 * characters (ASCII 96).
 *
 * Another way of phrasing this is to say that the encoded 6 bit characters are
 * mapped into the set:
 *
 *      `!"#$%&'()*+,-./012356789:;<=>?@ABC...XYZ[\]^_
 *
 * for transmission over communications lines.
 *
 * A small number of eight bit characters are encoded into a single line and a
 * count is put at the start of the line. Most lines in an encoded file have 45
 * encoded characters. When you look at a UU-encoded file note that most lines
 * start with the letter "M". "M" is decimal 77 which, minus the 32 bias, is
 * 45.
 *
 * Some encode programs put a check character at the end of each line. The
 * check is the sum of all the encoded characters, before adding the mapping,
 * modulo 64. Some encode programs have bugs in this line check routine; some
 * use alternative methods such as putting another line count character at the
 * end of a line or always ending a line with an "M". This library encodes
 * lines without line check characters. The decoder routine does not consider
 * line checks either.
 *
 * The lines of encoded data can be preceded by comments and by network
 * addressing information. The encoded data is directly preceded by a line
 * containing:
 *
 *      begin <file-mode> <file-name>
 *
 * For DOS, the file-mode has not much meaning. It will mostly be "644" for
 * standard acccess rights.
 *
 * The final end of encoded data is an encoded line with zero encoded
 * characters (a back-quote), followed by a line containing "end".
 *
 * To build the daemon, compile and link with DLIB.OBJ
 *
 *
 * This file is part of the Daemon server product.
 * Copyright (C) ITB CompuPhase
 * Permission is granted to use or modify this file. There is no warranty of
 * any kind.
 */

#include <stdio.h>
#include <string.h>
#include "daemon.h"
#include "uulib.h"


/*
 * If you are using Borland C++, the small or medium memory model and you need
 * a heap, you must set it explicitly. Otherwise the Daemon server assumes a
 * 64 kBytes heap. The compact, large and huge memory models use a different
 * heap strategy and do not use _heaplen.
 *
 * It is better to set the stack length as well. The stack length can be set in
 * any memory model.
 *
 * None of this applies to Microsoft C.
 */
#if defined(__TURBOC__)
  extern unsigned _stklen = 2048;       /* sufficient for this program */
  #if defined(__SMALL__) || defined(__MEDIUM__)
    extern unsigned _heaplen = 16;      /* set non-zero to avoid 64k heap */
  #endif
#endif


/*
 * We only export two functions. We add a third entry only to mark the end of
 * the list (this entry *must* be present).
 */
D_TABLE d_table = {
  { "uudecode", (D_EXPORTED) uudecode },
  { "uuencode", (D_EXPORTED) uuencode },
  { NULL, NULL} };


void main(void)
{
  /* check if wether we are running loaded by the Daemon Server */
  if (d_init() && d_inload())
    d_link(0, d_callback);              /* link to Daemon Server */

  /*
   * If we arrive here, we were not launched from the Daemon Server.
   */
  printf("UU-encode and decode daemon (does not run stand-alone)\n"
         "(c) Copyright 1995, ITB CompuPhase\n");
}


/* single character decode and encode */
#define DEC(c)  (((c) - ' ') & 0x3f)
#define ENC(c)  (char)(((c) & 0x3f) == 0 ? 0x60 : ((c) & 0x3f) + ' ')


int D_EXPORT uudecode(unsigned char _far *target, char _far *source)
{
  int len, retval;

  len = DEC(*source++);
  retval = len;
  while (len > 0) {
    if (len-- > 0)
      *target++ = (unsigned char)(( DEC(source[0]) << 2 ) | ( DEC(source[1]) >> 4 ));
    if (len-- > 0)
      *target++ = (unsigned char)(( DEC(source[1]) << 4 ) | ( DEC(source[2]) >> 2 ));
    if (len-- > 0)
      *target++ = (unsigned char)(( DEC(source[2]) << 6 ) | DEC(source[3]) );
    source += 4;
  } /* while */
  return retval;
}


int D_EXPORT uuencode(char _far *target, unsigned char _far *source,
                      int length)
{
  int split[4];

  if (length > 0x3f)
    return 0;                           /* can only encode multiples of 3 */

  *target++ = ENC(length);
  while (length > 0) {
    split[0] = source[0] >> 2;          /* split first byte to char. 0 & 1 */
    split[1] = source[0] << 4;
    if (length > 1) {
      split[1] |= source[1] >> 4;       /* split 2nd byte to char. 1 & 2 */
      split[2] = source[1] << 2;
      if (length > 2) {
        split[2] |= source[2] >> 6;     /* split 3th byte to char. 2 & 3 */
        split[3] = source[2];
      } /* if */
    } /* if */

    *target++ = ENC(split[0]);
    *target++ = ENC(split[1]);
    if (length > 1)
      *target++ = ENC(split[2]);
    if (length > 2)
      *target++ = ENC(split[3]);
    source += 3;
    length -= 3;
  } /* while */

  *target = '\0';                       /* end string */
  return 1;
}

